# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = "0.4.1.dev20+gc4aab1a.d20230829"
__version_tuple__ = version_tuple = (0, 4, 1, "dev20", "gc4aab1a.d20230829")
